from importlib.metadata import version

from ._core.merge import concat
from ._core.read import read_h5ad, read_h5td, read_zarr
from ._core.treedata import TreeData

__version__ = version("treedata")



from __future__ import annotations

import warnings
from abc import abstractmethod
from collections import abc as cabc
from collections import defaultdict
from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from functools import reduce
from typing import TYPE_CHECKING

import anndata as ad
import networkx as nx
import numpy as np
import pandas as pd
from scipy import sparse

from treedata._utils import subset_tree

if TYPE_CHECKING:
    from anndata import AnnData
    from anndata.raw import Raw

    from treedata._core.treedata import TreeData

    Index1D = slice | int | str | np.int64 | np.ndarray | list[str | int] | pd.Index
    Index = Index1D | tuple[Index1D | slice, Index1D | slice] | sparse.spmatrix | sparse.sparray


class AxisTreesBase(cabc.MutableMapping):
    """Mapping of key to nx.DiGraph aligned to an axis of parent TreeData."""

    @abstractmethod
    def __init__(self):
        """Abstract constructor."""
        self._tree_to_node = defaultdict(set)
        self._node_to_tree = defaultdict(set)
        self._axis = 0
        self._parent = None
        self._dimnames = ("obs", "var")

    def __repr__(self) -> str:
        """String representation of the object."""
        return f"{type(self).__name__} with keys: {', '.join(self.keys())}"

    def _ipython_key_completions_(self) -> list[str]:
        """IPython key completions."""
        return list(self.keys())

    def _validate_tree(self, tree: nx.DiGraph, key: str) -> set[str]:
        """Validate that graph is a tree and check for overlaps."""
        # Check value type
        if not isinstance(tree, nx.DiGraph):
            raise ValueError(f"Value for key {key} must be a nx.DiGraph")
        # Empty tree
        if tree.number_of_nodes() == 0:
            return set()
        # Check tree topology
        if not nx.is_tree(tree):
            raise ValueError(f"Value for key {key} must be a tree")
        # Check alignment
        nodes = set(tree.nodes)
        if self.parent.alignment == "leaves":
            nodes = {node for node in nodes if tree.out_degree(node) == 0}
        if self.parent.alignment in ["leaves", "nodes"]:
            if not nodes.issubset(self.dim_names):
                raise ValueError(
                    f"Names of {self.parent.alignment} must be in {self.dim}_names when alignment='{self.parent.alignment}'"
                )
        # Check overlap
        if not self.parent.allow_overlap:
            if key in self._tree_to_node:
                new_nodes = nodes.difference(self._tree_to_node[key])
            else:
                new_nodes = nodes
            if self._check_tree_overlap(new_nodes):
                raise ValueError(
                    "Leaf names overlap with leaf names of other trees. Set `allow_overlap=True` to allow this."
                )
        return nodes

    def _check_tree_overlap(self, nodes=()) -> bool:
        """Check if the leaves overlap with other trees."""
        nodes = set(nodes)
        for _, tree_nodes in self._tree_to_node.items():
            if nodes.intersection(tree_nodes):
                return True
            else:
                nodes = nodes.union(tree_nodes)
        return False

    def _update_tree_labels(self):
        """Update the tree labels in the parent object."""
        if self.parent._tree_label is not None:
            if self.parent.allow_overlap:
                labels = {k: ",".join(map(str, sorted(v))) for k, v in self._node_to_tree.items()}
            else:
                labels = {k: next(iter(v)) for k, v in self._node_to_tree.items()}
            getattr(self.parent, self.dim)[self.parent._tree_label] = getattr(self.parent, f"{self.dim}_names").map(
                labels
            )

    def _check_uniqueness(self):
        """Check if the names of the axis are unique."""
        names = "Observation" if self.dim == "obs" else "Variable"
        if not getattr(self.parent, self.dim).index.is_unique:
            warnings.warn(
                f"{names} names must be unique to store a tree. Calling `.{self.dim}_names_make_unique` to make them unique.",
                stacklevel=2,
            )
            getattr(self.parent, self.dim).index = ad.utils.make_index_unique(getattr(self.parent, self.dim).index)

    def copy(self) -> AxisTrees:
        """Returns a deep copy of the object."""
        d = AxisTrees(self.parent, self._axis)
        for k, v in self.items():
            d[k] = v.copy()
        return d

    def _view(self, parent: TreeData, subset_idx: Index1D | None) -> AxisTreesView:
        """Returns a subset copy-on-write view of the object."""
        return AxisTreesView(self, parent, subset_idx)

    @property
    def parent(self) -> AnnData | Raw:
        """Parent object of the mapping."""
        return self._parent

    @property
    def attrname(self) -> str:
        """Name of the attribute this is aligned to."""
        return f"{self.dim}t"

    @property
    def axes(self) -> tuple[int]:
        """Axes of the parent this is aligned to"""
        return (self._axis,)

    @property
    def dim(self) -> str:
        """Name of the dimension this aligned to."""
        return self._dimnames[self._axis]

    @property
    def dim_names(self) -> pd.Index:
        """Names of the dimension this aligned to."""
        return (self.parent.obs_names, self.parent.var_names)[self._axis]


class AxisTrees(AxisTreesBase):
    """Mapping of key to nx.DiGraph aligned to an axis of parent TreeData."""

    def __init__(
        self,
        parent: TreeData,
        axis: int,
        vals: Mapping | None = None,
    ):
        self._parent = parent
        self._dimnames = ("obs", "var")
        if axis not in (0, 1):
            raise ValueError()
        self._axis = axis
        self._data = {}
        self._tree_to_node = defaultdict(set)
        self._node_to_tree = defaultdict(set)
        if vals is not None:
            self.update(vals)

    def __getitem__(self, key: str) -> nx.DiGraph:
        """Get item from the mapping."""
        return nx.graphviews.generic_graph_view(self._data[key])

    def __setitem__(self, key: str, value: nx.DiGraph):
        """Set item in the mapping."""
        self._check_uniqueness()
        nodes = self._validate_tree(value, key)

        for node in nodes:
            self._node_to_tree[node].add(key)
        self._tree_to_node[key] = nodes

        if not self.parent.is_view:
            self._update_tree_labels()

        self._data[key] = value.copy()

    def __delitem__(self, key: str):
        """Delete item from the mapping."""
        for leaf in self._tree_to_node[key]:
            self._node_to_tree[leaf].remove(key)
            if not self._node_to_tree[leaf]:
                del self._node_to_tree[leaf]
        del self._tree_to_node[key]

        self._update_tree_labels()

        del self._data[key]

    def __len__(self) -> int:
        """Get length of the mapping."""
        return len(self._data)

    def __contains__(self, key: str) -> bool:
        """Check if the mapping contains a key."""
        return key in self._data

    def __iter__(self) -> Iterator[str]:
        """Iterate over the keys of the mapping."""
        return iter(self._data)

    def _validate_mapping(self) -> bool:
        """Check that the mapping is valid."""
        if self.parent.alignment in ["subset"]:
            return True
        for key, value in self.items():
            try:
                self._validate_tree(value, key)
            except ValueError:
                return False
        return True


class AxisTreesView(AxisTreesBase):
    """View of AxisTree object."""

    def __init__(
        self,
        parent_trees: AxisTreesBase,
        parent_view: TreeData,
        subset_idx: Index1D | None,
    ):
        self.parent_trees = parent_trees
        self._parent = parent_view
        self._dimnames = ("obs", "var")
        self.subset_idx = subset_idx
        self._axis = parent_trees._axis
        self._tree_to_node = parent_trees._tree_to_node
        self._node_to_tree = parent_trees._node_to_tree

    def __getitem__(self, key: str) -> nx.DiGraph:
        """Get item from the mapping."""
        if self.parent.alignment in ["leaves", "nodes"]:
            nodes = self.parent_trees._tree_to_node[key]
            subset_nodes = nodes.intersection(self.dim_names.values)
            tree_view = subset_tree(self.parent_trees[key], subset_nodes, asview=True, alignment=self.parent.alignment)
        else:
            tree_view = nx.graphviews.generic_graph_view(self.parent_trees[key])
        return tree_view

    def __setitem__(self, key: str, value: nx.DiGraph):
        """Set item in the mapping."""
        self._check_uniqueness()
        self._validate_tree(value, key)
        warnings.warn(
            f"Setting element `.{self.attrname}['{key}']` of view, initializing view as actual.", stacklevel=2
        )
        with view_update(self.parent, self.attrname, ()) as new_mapping:
            new_mapping[key] = value

    def __delitem__(self, key: str):
        """Delete item from the mapping."""
        if key not in self:
            raise KeyError(
                "'{key!r}' not found in view of {self.attrname}"
            )  # Make sure it exists before bothering with a copy
        warnings.warn(
            f"Removing element `.{self.attrname}['{key}']` of view, initializing view as actual.", stacklevel=2
        )
        with view_update(self.parent, self.attrname, ()) as new_mapping:
            del new_mapping[key]

    def __iter__(self) -> Iterator[str]:
        """Iterate over the keys of the mapping."""
        return iter(self.parent_trees)

    def __contains__(self, key: str) -> bool:
        """Check if the mapping contains a key."""
        return key in self.parent_trees

    def __len__(self) -> int:
        """Get length of the mapping."""
        return len(self.parent_trees)

    def _validate_mapping(self) -> bool:
        """Check that the mapping is valid."""
        if self.parent.alignment in ["leaves", "subset"]:
            return True
        for key in self.keys():
            try:
                self[key]
            except ValueError:
                return False
        return True


@contextmanager
def view_update(tdata_view: TreeData, attr_name: str, keys: tuple[str, ...]) -> Iterator:
    """Context manager for updating a view of an TreeData object.

    Contains logic for "actualizing" a view. Yields the object to be modified in-place.

    Parameters
    ----------
    tree_view
        A view of a TreeData object
    attr_name
        Name of the attribute being updated
    keys
        Keys to the attribute being updated

    Yields
    ------
    `adata.attr[key1][key2][keyn]...`
    """
    new = tdata_view.copy()
    attr = getattr(new, attr_name)
    container = reduce(lambda d, k: d[k], keys, attr)
    yield container
    tdata_view._init_as_actual(new)



from __future__ import annotations

import json
import warnings
from collections.abc import MutableMapping
from os import PathLike
from pathlib import Path
from typing import Any, Literal

import anndata as ad
import h5py
import networkx as nx
import numpy as np
import pandas as pd
import zarr
from packaging import version

from treedata._core.aligned_mapping import AxisTrees
from treedata._core.treedata import TreeData

ANDATA_VERSION = version.parse(ad.__version__)
USE_EXPERIMENTAL = ANDATA_VERSION < version.parse("0.11.0")


def _make_serializable(data: Any) -> Any:
    """Make a dictionary serializable."""
    if isinstance(data, dict):
        return {k: _make_serializable(v) for k, v in data.items()}
    elif isinstance(data, list | tuple | set):
        return [_make_serializable(v) for v in data]
    elif isinstance(data, np.ndarray):
        return data.tolist()
    elif isinstance(data, np.generic | np.number):
        return data.item()
    elif isinstance(data, pd.Series):
        return data.tolist()
    else:
        return data


def _write_elem(f, k, elem, *, dataset_kwargs) -> None:
    """Write an element to a storage group using anndata encoding."""
    if USE_EXPERIMENTAL:
        ad.experimental.write_elem(f, k, elem, dataset_kwargs=dataset_kwargs)
    else:
        ad.io.write_elem(f, k, elem, dataset_kwargs=dataset_kwargs)


def _digraph_to_dict(G: nx.DiGraph) -> dict:
    """Convert a networkx.DiGraph to a dictionary."""
    G = nx.DiGraph(G)
    edge_dict = nx.to_dict_of_dicts(G)
    # Get node data
    node_dict = {node: G.nodes[node] for node in G.nodes()}
    # Combine edge and node data in one dictionary
    graph_dict = {"edges": edge_dict, "nodes": node_dict}
    return graph_dict


def _serialize_axis_trees(trees: AxisTrees) -> str:
    """Serialize AxisTrees."""
    d = {k: _digraph_to_dict(v) for k, v in trees.items()}
    return json.dumps(_make_serializable(d))


def _write_tdata(f, tdata, filename, **kwargs) -> None:
    """Write TreeData to file."""
    # Add encoding type and version
    f = f["/"]
    f.attrs.setdefault("encoding-type", "treedata")
    f.attrs.setdefault("encoding-version", "0.1.0")
    # Convert strings to categoricals
    tdata.strings_to_categoricals()
    # Write X if not backed
    if not (tdata.isbacked and Path(tdata.filename) == Path(filename)):
        _write_elem(f, "X", tdata.X, dataset_kwargs=kwargs)
    # Write array elements
    for key in ["obs", "var", "label", "allow_overlap", "alignment"]:
        _write_elem(f, key, getattr(tdata, key), dataset_kwargs=kwargs)
    # Write group elements
    for key in ["obsm", "varm", "obsp", "varp", "layers", "uns"]:
        _write_elem(f, key, dict(getattr(tdata, key)), dataset_kwargs=kwargs)
    # Write axis tree elements
    for key in ["obst", "vart"]:
        _write_elem(f, key, _serialize_axis_trees(getattr(tdata, key)), dataset_kwargs=kwargs)
    # Write raw
    if tdata.raw is not None:
        tdata.strings_to_categoricals(tdata.raw.var)
        _write_elem(f, "raw", tdata.raw, dataset_kwargs=kwargs)
    # Close the file
    tdata.file.close()


def write_h5td(
    filename: PathLike | None,
    tdata: TreeData,
    compression: Literal["gzip", "lzf"] | None = None,
    compression_opts: int | Any = None,
    **kwargs,
) -> None:
    """Write `.h5td`-formatted hdf5 file.

    Parameters
    ----------
    filename
        Filename of data file. Defaults to backing file.
    tdata
        TreeData object to write.
    compression
        [`lzf`, `gzip`], see the h5py :ref:`dataset_compression`.
    compression_opts
        [`lzf`, `gzip`], see the h5py :ref:`dataset_compression`.
    """
    mode = "a" if tdata.isbacked else "w"
    if tdata.isbacked:  # close so that we can reopen below
        tdata.file.close()
    with h5py.File(filename, mode) as f:
        _write_tdata(f, tdata, filename, compression=compression, compression_opts=compression_opts, **kwargs)


def write_h5ad(
    filename: PathLike | None,
    tdata: TreeData,
    compression: Literal["gzip", "lzf"] | None = None,
    compression_opts: int | Any = None,
    **kwargs,
) -> None:
    """Write `.h5td`-formatted hdf5 file. Deprecated, use `write_h5td` instead.

    Parameters
    ----------
    filename
        Filename of data file. Defaults to backing file.
    tdata
        TreeData object to write.
    compression
        [`lzf`, `gzip`], see the h5py :ref:`dataset_compression`.
    compression_opts
        [`lzf`, `gzip`], see the h5py :ref:`dataset_compression`.
    """
    warnings.warn(
        "write_h5ad has been renamed to write_h5td. write_h5ad will be removed in v1.0.0.",
        DeprecationWarning,
        stacklevel=2,
    )
    write_h5td(
        filename=filename,
        tdata=tdata,
        compression=compression,
        compression_opts=compression_opts,
        **kwargs,
    )


def write_zarr(filename: MutableMapping | PathLike, tdata: TreeData, **kwargs) -> None:
    """Write `.zarr`-formatted zarr file.

    Parameters
    ----------
    filename
        Filename of data file. Defaults to backing file.
    tdata
        TreeData object to write.
    kwargs
        Additional keyword arguments passed to :func:`zarr.save`.
    """
    with zarr.open(filename, mode="w") as f:
        _write_tdata(f, tdata, filename, **kwargs)



import networkx as nx
import numpy as np
import pandas as pd
import pytest

import treedata as td


@pytest.fixture
def tree():
    tree = nx.balanced_tree(r=2, h=3, create_using=nx.DiGraph)
    tree = nx.relabel_nodes(tree, {i: str(i) for i in tree.nodes})
    depths = nx.single_source_shortest_path_length(tree, "0")
    nx.set_node_attributes(tree, values=depths, name="depth")
    yield tree


@pytest.fixture
def tdata(tree):
    df = pd.DataFrame({"anno": range(8)}, index=[str(i) for i in range(7, 15)])
    yield td.TreeData(
        X=np.zeros((8, 8)), obst={"0": tree}, vart={"0": tree}, obs=df, var=df, allow_overlap=True, label="tree"
    )


@pytest.fixture
def nodes_tdata(tree):
    df = pd.DataFrame({"anno": range(15)}, index=[str(i) for i in range(15)])
    yield td.TreeData(
        X=np.zeros((15, 15)),
        obst={"tree": tree},
        vart={"tree": tree},
        obs=df,
        var=df,
        alignment="nodes",
        allow_overlap=True,
    )


@pytest.fixture
def tdata_list(tdata):
    other_tree = nx.DiGraph()
    other_tree.add_edges_from([("0", "7"), ("0", "8")])
    tdata_1 = tdata[:2, :].copy()
    tdata_1.obst["1"] = other_tree
    tdata_1.vart["1"] = other_tree
    yield [tdata_1, tdata[2:4, :].copy(), tdata[4:, :4].copy()]


def test_concat(tdata_list):
    # outer join
    tdata = td.concat(tdata_list, axis="obs", label="subset", join="outer")
    assert list(tdata.obs["subset"]) == ["0"] * 2 + ["1"] * 2 + ["2"] * 4
    assert tdata.obst["0"].number_of_nodes() == 15
    assert tdata.obst["1"].number_of_nodes() == 3
    assert tdata.shape == (8, 8)
    # inner join
    tdata = td.concat(tdata_list, axis=0, label="subset", join="inner")
    assert list(tdata.obs["subset"]) == ["0"] * 2 + ["1"] * 2 + ["2"] * 4
    assert tdata.shape == (8, 4)
    # test join dictonary
    tdata_dict = {str(i): tdata for i, tdata in enumerate(tdata_list)}
    tdata = td.concat(tdata_dict, axis=0, label="subset", join="inner")
    assert tdata.shape == (8, 4)


def test_merge_outer(tdata_list):
    # None
    tdata = td.concat(tdata_list, axis=0, join="outer", merge=None)
    assert list(tdata.vart.keys()) == []
    # same
    tdata = td.concat(tdata_list, axis=0, join="outer", merge="same")
    assert list(tdata.vart.keys()) == []
    # unique
    tdata = td.concat(tdata_list, axis=0, join="outer", merge="first")
    assert list(tdata.vart.keys()) == ["0", "1"]
    # only
    tdata = td.concat(tdata_list, axis=0, join="outer", merge="only")
    assert list(tdata.vart.keys()) == ["1"]
    # first
    tdata = td.concat(tdata_list, axis=0, join="outer", merge="first")
    assert list(tdata.vart.keys()) == ["0", "1"]
    assert tdata.vart["0"].number_of_nodes() == 15
    assert tdata.vart["1"].number_of_nodes() == 3


def test_merge_inner(tdata_list):
    # None
    tdata = td.concat(tdata_list, axis=0, join="inner", merge=None)
    assert list(tdata.vart.keys()) == []
    # same
    tdata = td.concat(tdata_list, axis=0, join="inner", merge="same")
    assert list(tdata.vart.keys()) == ["0"]
    # unique
    tdata = td.concat(tdata_list, axis=0, join="inner", merge="first")
    assert list(tdata.vart.keys()) == ["0", "1"]
    # only
    tdata = td.concat(tdata_list, axis=0, join="inner", merge="only")
    assert list(tdata.vart.keys()) == ["1"]
    # first
    tdata = td.concat(tdata_list, axis=0, join="inner", merge="first")
    assert list(tdata.vart.keys()) == ["0", "1"]
    assert tdata.vart["0"].number_of_nodes() == 8
    assert tdata.vart["1"].number_of_nodes() == 3


def test_node_alignment(nodes_tdata):
    tdata1 = nodes_tdata[:14, :].copy()
    tdata2 = nodes_tdata[14:, :].copy()
    assert tdata1.alignment == "nodes"
    assert tdata2.alignment == "nodes"
    # concat same obst key
    with pytest.warns(UserWarning):
        tdata = td.concat([tdata1, tdata2], axis=0, join="outer", merge="same")
    assert tdata.shape == (15, 15)
    print(tdata)
    assert tdata.alignment == "nodes"
    assert list(tdata.obst.keys()) == []
    assert list(tdata.vart.keys()) == ["tree"]
    # concat different obst keys
    tdata2.obst["tree2"] = tdata2.obst.pop("tree")
    tdata = td.concat([tdata1, tdata2], axis=0, join="outer", merge="same")
    assert tdata.shape == (15, 15)
    assert tdata.alignment == "nodes"
    assert set(tdata.obst.keys()) == {"tree", "tree2"}
    assert list(tdata.vart.keys()) == ["tree"]
    assert tdata.obst["tree"].number_of_nodes() == 14
    assert tdata.obst["tree2"].number_of_nodes() == 1


def test_subset_alignment(nodes_tdata):
    with pytest.warns(UserWarning):
        tdata1 = nodes_tdata[[0, 2, 4, 6, 8, 10, 12, 14], :].copy()
    with pytest.warns(UserWarning):
        tdata2 = nodes_tdata[[1, 3, 5, 7, 9, 11, 13], :].copy()
    assert tdata1.alignment == "subset"
    assert tdata2.alignment == "subset"
    # concat same obst key
    with pytest.warns(UserWarning):
        tdata = td.concat([tdata1, tdata2], axis=0, join="outer", merge="same")
    assert tdata.shape == (15, 15)
    assert tdata.alignment == "subset"
    assert list(tdata.obst.keys()) == ["tree"]
    assert list(tdata.vart.keys()) == ["tree"]


def test_concat_bad_index(tdata_list):
    tdata_list[0].obs.index = tdata_list[1].obs.index
    with pytest.raises(ValueError):
        td.concat(tdata_list, axis=0, join="outer")


def test_concat_bad_tree(tdata_list):
    bad_tree = nx.DiGraph()
    bad_tree.add_edges_from([("bad", "7")])
    tdata_list[0].obst["0"] = bad_tree
    with pytest.raises(ValueError):
        td.concat(tdata_list, axis=0, join="outer")


if __name__ == "__main__":
    pytest.main(["-v", __file__])



from collections import deque
from typing import Literal

import networkx as nx


def subset_tree(
    tree: nx.DiGraph, nodes: list[str | int] | set[str | int], asview: bool, alignment: str = "leaves"
) -> nx.DiGraph:
    """Subset tree."""
    keep_nodes = set(nodes)
    # if leaves add all ancestors to keep_nodes
    if alignment == "leaves":
        nodes_to_check = deque()
        for node in nodes:
            nodes_to_check.extend(tree.predecessors(node))
        while nodes_to_check:
            node = nodes_to_check.popleft()
            if node in keep_nodes:
                continue
            else:
                keep_nodes.add(node)
                nodes_to_check.extend(tree.predecessors(node))
    subgraph = tree.subgraph(keep_nodes)
    if len(keep_nodes) > 0 and not nx.is_tree(subgraph):
        raise ValueError("Subset is not a tree. Please check your input.")
    if asview:
        return tree.subgraph(keep_nodes)
    else:
        return tree.subgraph(keep_nodes).copy()


def combine_trees(subsets: list[nx.DiGraph]) -> nx.DiGraph:
    """Combine two or more subsets of a tree into a single tree."""
    # Initialize a new directed graph for the combined tree
    combined_tree = nx.DiGraph()

    # Iterate through each subset and add its nodes and edges to the combined tree
    for subset in subsets:
        combined_tree.add_nodes_from(subset.nodes(data=True))
        combined_tree.add_edges_from(subset.edges(data=True))

    # The combined_tree now contains all nodes and edges from the subsets
    return combined_tree


def _resolve_axis(
    axis: Literal["obs", 0, "var", 1],
) -> tuple[Literal[0], Literal["obs"]] | tuple[Literal[1], Literal["var"]]:
    """Resolve axis argument."""
    if axis in {0, "obs"}:
        return (0, "obs")
    if axis in {1, "var"}:
        return (1, "var")
    raise ValueError(f"`axis` must be either 0, 1, 'obs', or 'var', was {axis}")
